@@ -89,6 +89,8 @@ static struct fuse_opt xrootdfs_opts[14];
 
 enum { OPT_KEY_HELP, OPT_KEY_SECSSS, };
 
+bool usingEC = false;
+
 static void* xrootdfs_init(struct fuse_conn_info *conn)
 {
     struct passwd pw, *pwp;
@@ -130,6 +132,19 @@ static void* xrootdfs_init(struct fuse_conn_info *conn)
     XrdPosixXrootd *abc = new XrdPosixXrootd(-xrootdfs.maxfd);
     XrdFfsMisc_xrd_init(xrootdfs.rdr,xrootdfs.urlcachelife,0);
     XrdFfsWcache_init(abc->fdOrigin(), xrootdfs.maxfd);
+
+    char *next, *savptr;
+    next = strtok_r(strdup(xrootdfs.rdr), "//", &savptr);
+    next = strtok_r(NULL, "//", &savptr);
+    char exportpath[1024];
+    while ((next = strtok_r(NULL, "//", &savptr)) != NULL)
+    {
+        strcat(exportpath, "/");
+        strcat(exportpath, next);
+    }
+    setenv("XRDEXPORTS", exportpath, 1);
+    if (savptr) free(savptr);
+    if (next) free(next);
 /*
    From FAQ:
       Miscellaneous threads should be started from the init() method.
@@ -442,10 +457,13 @@ static int xrootdfs_create(const char *path, mode_t mode, struct fuse_file_info
     int res, fd;
     if (!S_ISREG(mode))
         return -EPERM;
-    res = xrootdfs_do_create(path, xrootdfs.rdr, O_CREAT | O_WRONLY, true, &fd);
+    if (usingEC)
+        res = xrootdfs_do_create(path, xrootdfs.rdr, O_CREAT | O_WRONLY | O_EXCL, true, &fd);
+    else
+        res = xrootdfs_do_create(path, xrootdfs.rdr, O_CREAT | O_WRONLY, true, &fd);
     if (res < 0) return res;
     fi->fh = fd;
-    XrdFfsWcache_create(fd);    // Unlike mknod and like open, prepare wcache.
+    XrdFfsWcache_create(fd, fi->flags);    // Unlike mknod and like open, prepare wcache.
     if (xrootdfs.cns != NULL)
     {
         xrootdfs_do_create(path, xrootdfs.cns, O_CREAT | O_EXCL, false, &fd);
@@ -785,7 +803,7 @@ static int xrootdfs_open(const char *path, struct fuse_file_info *fi)
 
     fi->fh = fd;
     // be careful, 0 means error for this function
-    if (XrdFfsWcache_create(fi->fh))
+    if (XrdFfsWcache_create(fi->fh, fi->flags))
         return 0;
     else
         return -errno;
@@ -798,10 +816,31 @@ static int xrootdfs_read(const char *path, char *buf, size_t size, off_t offset,
     int res;
 
     fd = (int) fi->fh;
-    XrdFfsWcache_flush(fd);  /* in case is the file is reading/writing */
-    res = XrdFfsPosix_pread(fd, buf, size, offset);
+    if (fi->flags & O_RDWR) XrdFfsWcache_flush(fd);
+
+    if (usingEC)
+    {
+        struct stat stbuf;
+        XrdPosixXrootd::Fstat(fd, &stbuf);  // Silly but does not seem to hurt performance
+        off_t fsize = stbuf.st_size;
+
+        //!!! TO DO: Remove offset test once XrdClEC read regression is fixed 
+        if ( offset >= fsize )
+            return 0;
+
+        size = (size_t)(fsize - offset) > size ? size : fsize - offset; 
+        // Restrict the use of read cache to O_DIRECT use case 
+        // See comment in XRdFfsWcache_pread()
+        if ( ! (fi->flags & O_RDWR) && (fi->flags & O_DIRECT) )
+            res = XrdFfsWcache_pread(fd, buf, size, offset);
+        else
+            res = XrdFfsPosix_pread(fd, buf, size, offset);
+    }
+    else
+        res = XrdFfsPosix_pread(fd, buf, size, offset);
+
     if (res == -1)
-        res = -errno;
+    res = -errno;
 
     return res;
 }
@@ -1233,6 +1272,7 @@ static void xrootdfs_usage(const char *progname)
 "\n"
 "Default options:\n"
 "    fsname=xrootdfs,allow_other,max_write=131072,attr_timeout=10,entry_timeout=10,negative_timeout=5\n"
+"  In case of an Erasure Encoding storage, entry_timeout=0\n"
 "\n"
 "[Required]\n"
 "    -o rdr=redirector_url    root URL of the Xrootd redirector\n"
@@ -1308,13 +1348,25 @@ int main(int argc, char *argv[])
 /* Define XrootdFS options */
     char **cmdline_opts;
 
+    if (getenv("XRDCL_EC")) usingEC = true;
+
     cmdline_opts = (char **) malloc(sizeof(char*) * (argc -1 + 3));
     cmdline_opts[0] = argv[0];
     cmdline_opts[1] = strdup("-o");
     if (getenv("XROOTDFS_NO_ALLOW_OTHER") != NULL && ! strcmp(getenv("XROOTDFS_NO_ALLOW_OTHER"),"1") )
-        cmdline_opts[2] = strdup("fsname=xrootdfs,max_write=131072,attr_timeout=10,entry_timeout=10,negative_timeout=5");
+     {
+        if (! usingEC)
+            cmdline_opts[2] = strdup("fsname=xrootdfs,max_write=131072,attr_timeout=10,entry_timeout=10,negative_timeout=5");
+        else
+            cmdline_opts[2] = strdup("fsname=xrootdfs,max_write=131072,attr_timeout=10,entry_timeout=0,negative_timeout=5");
+    }
     else
-        cmdline_opts[2] = strdup("fsname=xrootdfs,allow_other,max_write=131072,attr_timeout=10,entry_timeout=10,negative_timeout=5");
+    {
+        if (! usingEC)
+            cmdline_opts[2] = strdup("fsname=xrootdfs,allow_other,max_write=131072,attr_timeout=10,entry_timeout=10,negative_timeout=5");
+        else
+            cmdline_opts[2] = strdup("fsname=xrootdfs,allow_other,max_write=131072,attr_timeout=10,entry_timeout=0,negative_timeout=5");
+    }
 
     for (int i = 1; i < argc; i++)
         cmdline_opts[i+2] = argv[i];
