@@ -90,6 +90,7 @@ char *XrdHttpProtocol::secretkey = 0;
 
 char *XrdHttpProtocol::gridmap = 0;
 bool XrdHttpProtocol::isRequiredGridmap = false;
+bool XrdHttpProtocol::compatNameGeneration = false;
 int XrdHttpProtocol::sslverifydepth = 9;
 BIO *XrdHttpProtocol::sslbio_err = 0;
 XrdHttpSecXtractor *XrdHttpProtocol::secxtractor = 0;
@@ -98,6 +99,8 @@ struct XrdHttpProtocol::XrdHttpExtHandlerInfo XrdHttpProtocol::exthandler[MAX_XR
 int XrdHttpProtocol::exthandlercnt = 0;
 std::map< std::string, std::string > XrdHttpProtocol::hdr2cgimap; 
 
+bool XrdHttpProtocol::usingEC = false;
+
 XrdScheduler *XrdHttpProtocol::Sched = 0; // System scheduler
 XrdBuffManager *XrdHttpProtocol::BPool = 0; // Buffer manager
 XrdSysError XrdHttpProtocol::eDest = 0; // Error message handler
@@ -1036,6 +1039,9 @@ int XrdHttpProtocol::Config(const char *ConfigFN, XrdOucEnv *myEnv) {
        return 1;
       }
 
+// Test if XrdEC is loaded
+   if (getenv("XRDCL_EC")) usingEC = true;
+
 // If https was disabled, then issue a warning message if xrdtls configured
 // of it's disabled because httpsmode was auto and xrdtls was not configured.
 // If we get past this point then we know https is a plausible option but we
@@ -1515,6 +1521,8 @@ int XrdHttpProtocol::StartSimpleResp(int code, const char *desc, const char *hea
   else
     ss << "Connection: Close" << crlf;
 
+  ss << "Server: XrootD/" << XrdVSTRING << crlf;
+  
   if ((bodylen >= 0) && (code != 100))
     ss << "Content-Length: " << bodylen << crlf;
 
@@ -1954,7 +1962,7 @@ int XrdHttpProtocol::xsslkey(XrdOucStream & Config) {
 
 /* Function: xgmap
 
-   Purpose:  To parse the directive: gridmap [required] <path>
+   Purpose:  To parse the directive: gridmap [required] [compatNameGeneration] <path>
 
      required   optional parameter which if present treats any grimap errors
                 as fatal.
@@ -1989,6 +1997,19 @@ int XrdHttpProtocol::xgmap(XrdOucStream & Config) {
     }
   }
 
+  // Handle optional parameter "compatNameGeneration"
+  //
+  if (!strcmp(val, "compatNameGeneration")) {
+    compatNameGeneration = true;
+    val = Config.GetWord();
+    if (!val || !val[0]) {
+      eDest.Emsg("Config", "HTTP X509 gridmap file missing after "
+                 "[compatNameGeneration] parameter");
+      return 1;
+    }
+  }
+
+
   // Record the path
   //
   if (gridmap) free(gridmap);
@@ -2426,7 +2447,7 @@ int XrdHttpProtocol::xsecxtractor(XrdOucStream& Config) {
     }
 
     char libName[4096];
-    strncpy(libName, val, sizeof(libName));
+    strlcpy(libName, val, sizeof(libName));
     libName[sizeof(libName) - 1] = '\0';
     char libParms[4096];
 
