@@ -43,7 +43,6 @@
 
 namespace XrdTlsGlobal
 {
-extern XrdTls::msgCB_t msgCB;
 extern XrdSysTrace SysTrace;
 };
   
@@ -134,7 +133,7 @@ do{ctxImpl->crlMutex.ReadLock();
 // Verify that the context was properly built
 //
    if (!newctx || !newctx->isOK())
-      {XrdTlsGlobal::msgCB("CrlRefresh:","Refresh of context failed!!!",false);
+      {XrdTls::Emsg("CrlRefresh:","Refresh of context failed!!!",false);
        continue;
       }
 
@@ -280,7 +279,7 @@ bool Setup_Flusher(XrdTlsContextImpl *pImpl, int flushT)
       {char eBuff[512];
        snprintf(eBuff, sizeof(eBuff),
                 "Unable to start cache flusher thread; rc=%d", rc);
-       XrdTlsGlobal::msgCB("SessCache:", eBuff, false);
+       XrdTls::Emsg("SessCache:", eBuff, false);
        return false;
       }
 
@@ -387,7 +386,9 @@ void InitTLS() // This is strictly a one-time call!
    OpenSSL_add_all_algorithms();
    SSL_load_error_strings();
    OpenSSL_add_all_ciphers();
+#if OPENSSL_VERSION_NUMBER < 0x30000000L
    ERR_load_BIO_strings();
+#endif
    ERR_load_crypto_strings();
 
 // Set callbacks if we need to do this
@@ -414,10 +415,10 @@ void Fatal(std::string *eMsg, const char *msg, bool sslmsg=false)
 //
    if (eMsg) *eMsg = msg;
 
-// Now route the message to the message callback function. We always flush the
-// ssl error queue, even if not an ssl error, to prevent suprises.
+// Now route the message to the message callback function. If this is an ssl
+// related error we also flush the ssl error queue to prevent suprises.
 //
-   XrdTlsGlobal::msgCB("TLS_Context:", msg, sslmsg);
+   XrdTls::Emsg("TLS_Context:", msg, sslmsg);
 }
 
 /******************************************************************************/
@@ -511,15 +512,15 @@ int VerCB(int aOK, X509_STORE_CTX *x509P)
 
        X509_NAME_oneline(X509_get_subject_name(cert), name, sizeof(name));
        snprintf(info,sizeof(info),"Cert verification failed for DN=%s",name);
-       XrdTlsGlobal::msgCB("CertVerify:", info, false);
+       XrdTls::Emsg("CertVerify:", info, false);
 
        X509_NAME_oneline(X509_get_issuer_name(cert), name, sizeof(name));
        snprintf(info,sizeof(info),"Failing cert issuer=%s", name);
-       XrdTlsGlobal::msgCB("CertVerify:", info, false);
+       XrdTls::Emsg("CertVerify:", info, false);
 
        snprintf(info, sizeof(info), "Error %d at depth %d [%s]", err, depth,
                                     X509_verify_cert_error_string(err));
-       XrdTlsGlobal::msgCB("CertVerify", info, false);
+       XrdTls::Emsg("CertVerify:", info, true);
       }
 
    return aOK;
@@ -533,6 +534,7 @@ int VerCB(int aOK, X509_STORE_CTX *x509P)
 /******************************************************************************/
 
 #define FATAL(msg) {Fatal(eMsg, msg); return;}
+#define FATAL_SSL(msg) {Fatal(eMsg, msg, true); return;}
   
 XrdTlsContext::XrdTlsContext(const char *cert,  const char *key,
                              const char *caDir, const char *caFile,
@@ -609,7 +611,7 @@ XrdTlsContext::XrdTlsContext(const char *cert,  const char *key,
 
 // Before we try to use any specified files, make sure they exist, are of
 // the right type and do not have excessive access privileges.
-//
+//                                                                              .a
    if (!VerPaths(cert, key, caDir, caFile, eText)) FATAL( eText.c_str());
 
 // Copy parameters to out parm structure.
@@ -633,7 +635,7 @@ XrdTlsContext::XrdTlsContext(const char *cert,  const char *key,
 
 // Make sure we have a context here
 //
-   if (pImpl->ctx == 0) FATAL("Unable to allocate TLS context!");
+   if (pImpl->ctx == 0) FATAL_SSL("Unable to allocate TLS context!");
 
 // Always prohibit SSLv2 & SSLv3 as these are not secure.
 //
@@ -653,7 +655,7 @@ XrdTlsContext::XrdTlsContext(const char *cert,  const char *key,
 //
    if (caDir || caFile)
      {if (!SSL_CTX_load_verify_locations(pImpl->ctx, caFile, caDir))
-         FATAL("Unable to load the CA cert file or directory.");
+         FATAL_SSL("Unable to load the CA cert file or directory.");
 
       int vDepth = (opts & vdept) >> vdepS;
       SSL_CTX_set_verify_depth(pImpl->ctx, (vDepth ? vDepth : 9));
@@ -674,7 +676,7 @@ XrdTlsContext::XrdTlsContext(const char *cert,  const char *key,
 // Set cipher list
 //
    if (!SSL_CTX_set_cipher_list(pImpl->ctx, sslCiphers))
-      FATAL("Unable to set SSL cipher list; no supported ciphers.");
+      FATAL_SSL("Unable to set SSL cipher list; no supported ciphers.");
 
 // If we need to enable eliptic-curve support, do so now. Note that for
 // OpenSSL 1.1.0+ this is automatically done for us.
@@ -704,17 +706,17 @@ XrdTlsContext::XrdTlsContext(const char *cert,  const char *key,
 // Load certificate
 //
    if (SSL_CTX_use_certificate_file(pImpl->ctx, cert, SSL_FILETYPE_PEM) != 1)
-      FATAL("Unable to create TLS context; invalid certificate.");
+      FATAL_SSL("Unable to create TLS context; invalid certificate.");
 
 // Load the private key
 //
    if (SSL_CTX_use_PrivateKey_file(pImpl->ctx, key, SSL_FILETYPE_PEM) != 1 )
-      FATAL("Unable to create TLS context; invalid private key.");
+      FATAL_SSL("Unable to create TLS context; invalid private key.");
 
 // Make sure the key and certificate file match.
 //
    if (SSL_CTX_check_private_key(pImpl->ctx) != 1 )
-      FATAL("Unable to create TLS context; cert-key mismatch.");
+      FATAL_SSL("Unable to create TLS context; cert-key mismatch.");
 
 // All went well, so keep the context.
 //
@@ -959,8 +961,12 @@ int XrdTlsContext::SessionCache(int opts, const char *id, int idlen)
 
 bool XrdTlsContext::SetContextCiphers(const char *ciphers)
 {
-   return (!(pImpl->ctx) || !SSL_CTX_set_cipher_list(pImpl->ctx, ciphers)
-         ? false : true);
+   if (pImpl->ctx && SSL_CTX_set_cipher_list(pImpl->ctx, ciphers)) return true;
+
+   char eBuff[2048];
+   snprintf(eBuff,sizeof(eBuff),"Unable to set context ciphers '%s'",ciphers);
+   Fatal(0, eBuff, true);
+   return false;
 }
 
 /******************************************************************************/
@@ -1005,8 +1011,8 @@ bool XrdTlsContext::SetCrlRefresh(int refsec)
 // Check if there is anything that is refreshable
 //
    if (!x509Verify())
-      {XrdTlsGlobal::msgCB("CrlRefresh:",
-                           "No cert information exists to refresh!", false);
+      {XrdTls::Emsg("CrlRefresh:",
+                    "No cert information exists to refresh!", false);
        return false;
       }
 
@@ -1024,7 +1030,7 @@ bool XrdTlsContext::SetCrlRefresh(int refsec)
           {char eBuff[512];
            snprintf(eBuff, sizeof(eBuff),
                     "Unable to start CRL refresh thread; rc=%d", rc);
-           XrdTlsGlobal::msgCB("CrlRefresh:", eBuff, false);
+           XrdTls::Emsg("CrlRefresh:", eBuff, false);
            pImpl->crlMutex.UnLock();
            return false;
           } else pImpl->crlRunning = true;
@@ -1040,8 +1046,8 @@ bool XrdTlsContext::SetCrlRefresh(int refsec)
 // Older version are too difficult to deal with. Issue a message if this
 // feature is being enabled on an old version.
 //
-   XrdTlsGlobal::msgCB("CrlRefresh:", "Refreshing CRLs only supported in "
-                       "OpenSSL version >= 1.02; CRL refresh disabled!", false);
+   XrdTls::Emsg("CrlRefresh:", "Refreshing CRLs only supported in "
+                "OpenSSL version >= 1.02; CRL refresh disabled!", false);
    return false;
 #endif
 }
